"""
Quick and dirty import of Hemp api's
"""
from hemp.tasks.stage import on, development, production, staging
from gitutils import remote_tags, last_remote_tag, clone, repo_url_to_project_name
from release import release_local
from workspace import create_temporary_workspace


from shutil import rmtree
from tempfile import mkdtemp

from git import Repo
from semver import bump_build, bump_prerelease, bump_patch, bump_major, bump_minor
from natsort.natsort import natsorted

from hemp.internal.utils import SimpleProgressPrinter, print_err, print_info, print_git_output


def release_local(url, version='patch', base='master', integration=None, default_version='0.0.1', use_prefix=None):
    # type: (str, str, str, str, str, str) -> str
    """
    Tag given repository with a new semver tag (bump version),
    optionally merging a integration branch.

    This will:
        - clone the repository to temporary directory
        - checkout branch indicated via base argument
        - retrieve all the tags, sort them in natural order
        - retrieve the last tag and bump it to given version
        - merge integration branch, if defined
        - tag and push base branch back to origin

    If no tag is present and version argument is any of the bump arguments,
    default_version will be used

    :rtype: str
    :param url: URL of the repository
    :param version: specific version or one of: build, prerelease, patch, minor, major
    :param base: base branch to use, by default master
    :param integration: integration branch to use, by default none
    :param default_version: default version used for when there are no tags and no specific version, default 0.0.1
    :param use_prefix: use prefix for tags - sometimes, 'v',
    :return: newly released version string or None if can not tag
    """
    workspace = mkdtemp()
    repo = Repo.clone_from(url, workspace, progress=SimpleProgressPrinter())

    if repo.bare:
        print_err('Cloned a bare repository, can not release [???]')

    origin = repo.remote('origin')

    if repo.active_branch.name != base:
        origin.fetch('refs/heads/{0}:refs/heads/{0}'.format(base), progress=SimpleProgressPrinter())
        repo.heads[base].checkout()

    if integration is not None and integration in repo.heads:
        print_info('Found integration branch "{0}", fetching'.format(integration))
        origin.fetch('refs/heads/{0}:refs/heads/{0}'.format(integration), progress=SimpleProgressPrinter())
        print_info('Will now attempt fast-forward {0} to include {1}'.format(base, integration))
        print_git_output(repo.git.merge('--commit', '--no-edit', '--stat', '--ff-only', '-v', integration))

    head_tags = (tag for tag in repo.tags if tag.commit == repo.head.commit)
    sorted_head_tags = natsorted(head_tags, key=lambda t: t.path)

    if 0 != len(sorted_head_tags):
        print_info(
            'Not going to auto-tag already tagged HEAD, tagged with {0}'
                .format(
                str.join(', ', (t.path[10:] for t in sorted_head_tags))
            )
        )
        return None

    last_tag = None

    if repo.tags:
        sorted_tags = natsorted(repo.tags, key=lambda t: t.path)
        current_tag = sorted_tags[-1].path[10:]
        print_info('Current tag is {0}'.format(current_tag))
        if use_prefix is not None and current_tag.startswith(use_prefix):
            last_tag = current_tag[len(use_prefix):]
        else:
            last_tag = current_tag

    print_info('Last known version: {0}'.format(last_tag))

    if last_tag is None:
        print('Unknown last version, using default version {0}'.format(default_version))
        last_tag = default_version

    if 'build' == version:
        next_version = bump_build(last_tag)

    elif 'prerelease' == version:
        next_version = bump_prerelease(last_tag)

    elif 'patch' == version:
        next_version = bump_patch(last_tag)

    elif 'minor' == version:
        next_version = bump_minor(last_tag)

    elif 'major' == version:
        next_version = bump_major(last_tag)
    else:
        next_version = version

    print_info('Next version: {0}'.format(next_version))

    next_tag = next_version.strip()

    if use_prefix is not None:
        next_tag = use_prefix + next_version

    print_info('Next tag: {0}'.format(next_tag))

    print_info('Tagging and pushing version')

    release_tag = repo.create_tag(
        path=next_tag,
        ref=repo.heads[base],
        message='Release tag of {0}'.format(next_version)
    )

    origin.push([release_tag, repo.heads[base]], progress=SimpleProgressPrinter())

    print_info('Done, clearing workspace')
    rmtree(workspace)

    return next_version



from os import chmod
from os import mkdir
from os.path import join
from tempfile import mkdtemp


def create_temporary_workspace(version=None, mode=0700):
    # type: (str, int) -> str
    """
    Create a temporary directory, optionally by placing a subdirectory: version
    :rtype: str
    :return: Directory path
    """
    workspace = mkdtemp('hemp_')
    if version is not None:
        workspace = join(workspace, version)
        mkdir(workspace, mode)
    chmod(workspace, mode)
    return workspace



from git import Git, Repo
from natsort.natsort import natsorted
from urlparse import urlparse

from hemp.internal.utils import SimpleProgressPrinter, print_info


def remote_tags(url):
    # type: (str) -> list
    """
    List all available remote tags naturally sorted as version strings
    :rtype: list
    :param url: Remote URL of the repository
    :return: list of available tags
    """
    tags = []
    remote_git = Git()
    for line in remote_git.ls_remote('--tags', '--quiet', url).split('\n'):
        hash_ref = line.split('\t')
        tags.append(hash_ref[1][10:].replace('^{}',''))
    return natsorted(tags)


def last_remote_tag(url):
    # type: (str) -> str
    """
    Get last tag from remote repository. This utility retrieves and sorts tags in natural order,
    not by date!
    :rtype: str
    :param url:
    :return: last available tag, sorted in natural order
    """
    return remote_tags(url)[-1]


def clone(url, directory, single_branch=None):
    print_info('Cloning {0} to {1} {2}'.format(
        url,
        directory,
        '[full clone]' if single_branch is None else '[{0}]'.format(single_branch)
    ))
    # type: (str, str, str) -> Repo
    """
    Clone a repository, optionally using shallow clone
    :rtype: Repo
    :param url: URL of the repository
    :param directory: Directory to clone to
    :param single_branch: branch to clone if shallow clone is preferred
    :return: GitPython repository object of the newly cloned repository
    """
    args = {
        'url': url,
        'to_path': directory,
        'progress': SimpleProgressPrinter(),
        'recursive': True
    }

    if single_branch is not None:
        args['depth'] = 1
        args['branch'] = single_branch
        args['single_branch'] = True

    return Repo.clone_from(**args)


def repo_url_to_project_name(url, normalize=True):
    # type: (str, bool) -> str
    """
    Extract project title from GIT repository URL, either https or ssh.
    Should work fine with github, bitbucket
    :rtype: str
    :param url: Url of the repository
    :param normalize: Title-case the project name
    :return: Project name
    """
    if url.startswith('http'):
        http_url = urlparse(url)
        name = http_url.path.replace('.git', '').lstrip('/').strip()
    else:
        name = url[url.index(':') + 1:].replace('.git', '').strip()
    if normalize:
        return name.replace('/', ' / ').title()
    return name



import sys

from fabric.utils import error, puts
from git import RemoteProgress


def print_err(message, func=None, exception=None, stdout=None, stderr=None):
    error('[Hemp] ' + message, func, exception, stdout, stderr)


def print_info(text, show_prefix=None, end="\n", flush=True):
    puts('[Hemp] ' + text, show_prefix, end, flush)


def print_git_output(stdout):
    for line in stdout.split('\n'):
        sys.stdout.write('[GIT] ' + line + '\n')
        sys.stdout.flush()


class SimpleProgressPrinter(RemoteProgress):
    def _parse_progress_line(self, line):
        if '\r' in line:
            line = line.replace('\r', '\r[GIT] ')
        sys.stdout.write('[GIT] ' + line + '\n')
        sys.stdout.flush()
